#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Program to test real and virtual memory subsystems.  This
 * program is capable of finding flaws in main memory and disk
 * subsystems if used correctly.
 *
 * Written by:  Chris Studholme
 * Last Update: 3-Nov-2003
 * Copyright:   GPL (http://www.fsf.org/copyleft/gpl.html)
 *
 * Theory of operation:  Four blocks of memory are allocated one
 * at a time.  Each block is filled with a sequence of numbers
 * that will be different in each block and essentially random.
 * Once four blocks have been allocated, the oldest block will
 * be checked to ensure that it still has the same sequence of
 * numbers and will then be deallocated.  A newly allocated
 * block will take its place.  Each block has a random size
 * varying from 1MB to MAX_BLOCK_SIZE.  
 *
 * To test both main memory and swap, set MAX_BLOCK_SIZE equal 
 * to half the amount of main memory you have.  To test only main
 * memory, set MAX_BLOCK_SIZE equal to one quarter your main memory.  
 * Some swapping will occur in this latter case.  
 *
 * This program can be very resource intensive.  Be patient with
 * your computer while it is running.  If you get an error indicating
 * that virual memory has been exhasted, reduce MAX_BLOCK_SIZE.
 * MAX_BLOCK_SIZE should never exceed (TOTAL_VIRTUAL_MEMORY/5).
 * 
 * Analysing the results:  After running the program for several
 * minutes or hours, if no errors are displayed, your memory is
 * fine.  If you do get errors, some things to look for are:
 *   - error is always in the same bit
 *   - error is always at the same offset from start of block
 *   - error is all over the place
 * When errors are reported, the size of each of the four blocks
 * is also reported.  If you find that errors only occur when
 * lots of memory are in use (more than physical RAM), the problem
 * may be in your disk subsystem.  If errors occur with equal 
 * frequency regardless of how much memory is in use, the problem
 * is most likely with your RAM chips or data bus.
 */

#define MAX_BLOCK_SIZE 32  /* MB */
#define SEC_PER_REPORT 60

#define INTS_PER_MB (1024*1024/sizeof(int))

int* blocks[4];
int blocksize[4];
int startvalue[4];

int main() {
  for(int i=0; i<4; ++i) 
    blocks[i]=0;
  
  int currentblock=0;
  time_t starttime=time(0), lasttime=time(0);
  long long totalMB=0;
  int errors=0;
  
  while(true) {
    // check current block
    if (blocks[currentblock]) 
      for (int i=0; i<blocksize[currentblock]; ++i) 
	if (blocks[currentblock][i]!=startvalue[currentblock]+i) {
	  fprintf(stderr,"Error at %.08X, %.08X should be %.08X,",
		  i,blocks[currentblock][i],startvalue[currentblock]+i);
	  fprintf(stderr," blocks are %d %d %d %d\n",
		  blocksize[currentblock]/INTS_PER_MB,
		  blocksize[(currentblock+1)&3]/INTS_PER_MB,
		  blocksize[(currentblock+2)&3]/INTS_PER_MB,
		  blocksize[(currentblock+3)&3]/INTS_PER_MB);
	  ++errors;
	}
    
    // free current block
    delete[] blocks[currentblock];
    blocks[currentblock]=0;
    
    // create new block
    blocksize[currentblock] = INTS_PER_MB * 
      (MAX_BLOCK_SIZE - (int)(drand48()*MAX_BLOCK_SIZE));
    blocks[currentblock] = new int[blocksize[currentblock]];
    if (blocks[currentblock]==0) {
      fprintf(stderr,"Cannot allocate %d, blocks are %d %d %d\n",
	      blocksize[currentblock]/INTS_PER_MB,
	      blocksize[(currentblock+1)&3]/INTS_PER_MB,
	      blocksize[(currentblock+2)&3]/INTS_PER_MB,
	      blocksize[(currentblock+3)&3]/INTS_PER_MB);
      blocksize[currentblock]=0;
    } else
      totalMB += blocksize[currentblock]/INTS_PER_MB;
    
    // fill new block
    startvalue[currentblock] = lrand48();
    for (int i=0; i<blocksize[currentblock]; ++i) 
      blocks[currentblock][i] = startvalue[currentblock]+i;
    
    // check if time for status report
    if (time(0) >= (lasttime+SEC_PER_REPORT)) {
      lasttime=time(0);
      char* s = ctime(&lasttime);
      s[24]=0;
      fprintf(stderr,"%s elapsed %ld sec, total %.1f GB tested, errors %d\n",s,
	      lasttime-starttime,(double)totalMB/1024,errors);
    }
    
    currentblock = (currentblock+1)&3;
  }
}
